%{
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sds.h"
#include "structs.h"
#include "symboltable.h"
#include "iascc.tab.h"

int yylex();
int yyparse();
FILE *yyin;

extern int yylineno;

void yyerror(const char *s);
void error(int line, const char *s);

void destroy_inst_operand(inst_operand* op);

/* format the given instruction and add it to the code array

type is one of INST_MOV...
*/
void append_inst(int type, inst_operand* op1, inst_operand* op2);

/* did we see an error during the last compilation, used
 * to decide if a .o file should be created */
int saw_error = 0;

/* add the label with the specified type to the symbol table,
 * reporting an error if it's already defined */
symbol_table_entry* create_symbol_table_entry(int line, sds label, symbol_table_entry_type type);

#define NUM_INST 16

/* boolean matrix of the possible addressing modes for
 * each of the operands of a given instruction */
int valid_addressing_per_inst[NUM_INST][2][NUM_ADDRESSING] = {
    /*   src           dst */
    {{1, 1, 1, 1}, {0, 1, 0, 1}}, /* mov */
    {{1, 1, 1, 1}, {1, 1, 1, 1}}, /* cmp */
    {{1, 1, 1, 1}, {0, 1, 0, 1}}, /* add */
    {{1, 1, 1, 1}, {0, 1, 0, 1}}, /* sub */
    {{0, 0, 0, 0}, {0, 1, 0, 1}}, /* not */
    {{0, 0, 0, 0}, {0, 1, 0, 1}}, /* clr */
    {{0, 1, 0, 0}, {0, 1, 0, 1}}, /* lea */
    {{0, 0, 0, 0}, {0, 1, 0, 1}}, /* inc */
    {{0, 0, 0, 0}, {0, 1, 0, 1}}, /* dec */
    {{0, 0, 0, 0}, {0, 1, 1, 1}}, /* jmp */
    {{0, 0, 0, 0}, {0, 1, 1, 1}}, /* bne */
    {{0, 0, 0, 0}, {0, 1, 1, 1}}, /* red */
    {{0, 0, 0, 0}, {1, 1, 1, 1}}, /* prn */
    {{0, 0, 0, 0}, {0, 1, 0, 0}}, /* jsr */
    {{0, 0, 0, 0}, {0, 0, 0, 0}}, /* rts */
    {{0, 0, 0, 0}, {0, 0, 0, 0}}, /* stop */
};

/* string representation of an instruction, for
 * error messages */
const char* instructions_string[NUM_INST] = {
    "mov",
    "cmp",
    "add",
    "sub",
    "not",
    "clr",
    "lea",
    "inc",
    "dec",
    "jmp",
    "bne",
    "red",
    "prn",
    "jsr",
    "rts",
    "stop"
};

/* convert a %token to zero based notation for usage with
 * valid_addressing_per_inst */
int insttoken_to_zero_index(int token) {
    int first = INST_MOV;
    assert(token - first >= 0);

    return token - first;
}

/* helper to convert the given addressing to its
 * string representation, for error messages */
const char* addressing_to_string[NUM_ADDRESSING] = {
    "immediate",
    "direct",
    "dist",
    "iregister"
};

/* verify that the given operand's addressing is valid for
 * the given instruction */
int check_operand_addressing(int line, int insttoken, int issrc, inst_operand op) {
    int zerobased_token = insttoken_to_zero_index(insttoken);
    int addressing_legal = valid_addressing_per_inst[zerobased_token][issrc ? 0 : 1][op.kind];

    if (!addressing_legal) {
        sds s = sdscatprintf(sdsempty(),
                             "invalid addressing %s for instruction %s",
                             addressing_to_string[op.kind],
                             instructions_string[zerobased_token]);
        error(line, s);
        sdsfree(s);

        return 0;
    }

    return 1;
}

%}
%debug
%locations

%union {
	int ival;
	sds sval;
    int_array* ints;
    inst_operand operand;
}

%token <ival> COMMENT

%token DIRECTIVE_DATA DIRECTIVE_STRING DIRECTIVE_ENTRY DIRECTIVE_EXTERN
%token <ival> INST_MOV INST_CMP INST_ADD INST_SUB
%token <ival> INST_NOT INST_CLR INST_LEA INST_INC INST_DEC INST_JMP INST_BNE INST_RED INST_PRN INST_JSR
%token <ival> INST_RTS INST_STOP

%token <ival> INT OP_INT REGISTER
%token <sval> STRING
%token <sval> LABEL OP_LABEL

%type <ints> numbers
%type <sval> maybe_label
%type <operand> op_operand
%type <ival> instructions_0_operand instructions_1_operand instructions_2_operands
%%

/* the first rule defined is the highest-level rule */
asfile: statements
      ;

statements: /* empty */
          | statements '\n'
          | statements directive_statement
          | statements op_statement
          | statements comment_statement
          | statements error '\n'
{
    error(@2.first_line, "unrecognized statement, expected one of: "
                         "comment, .data, .string, instruction");
}
;

comment_statement: COMMENT '\n'
                 ;

maybe_label: /* empty */
{
    $$ = NULL;
}
           | LABEL
{
    $$ = $1;
}
           ;

instructions_2_operands: INST_MOV
                       | INST_CMP
                       | INST_ADD
                       | INST_SUB
                       | INST_LEA
                       ;

instructions_1_operand: INST_NOT
                      | INST_CLR
                      | INST_INC
                      | INST_DEC
                      | INST_JMP
                      | INST_BNE
                      | INST_RED
                      | INST_PRN
                      | INST_JSR
                      ;

instructions_0_operand: INST_RTS
                      | INST_STOP
                      ;

/********** code generation **********

   1) create an entry in the symbol table if we're given a label
   2) check the addressing of the operands
   3) add the instruction to the code array

*/
op_statement: maybe_label instructions_2_operands op_operand ',' op_operand '\n'
{
    int inst = $2;
    inst_operand src = $3;
    inst_operand dst = $5;

    if ($1 != NULL)
        create_symbol_table_entry(@1.first_line, $1, entry_code);

    if (!check_operand_addressing(@3.first_line, inst, 1, src) || !check_operand_addressing(@5.first_line, inst, 0, dst)) {
        ;
    }
    else {
        append_inst(inst, &src, &dst);
    }

    destroy_inst_operand(&src);
    destroy_inst_operand(&dst);
    sdsfree($1);
}
            | maybe_label instructions_1_operand op_operand '\n'
{
    int inst = $2;
    inst_operand dst = $3;

    if ($1 != NULL)
        create_symbol_table_entry(@1.first_line, $1, entry_code);

    if (!check_operand_addressing(@3.first_line, inst, 0, dst)) {
        ;
    }
    else {
        append_inst(inst, &dst, NULL);
    }

    destroy_inst_operand(&dst);
    sdsfree($1);
}
            | maybe_label instructions_0_operand '\n'
{
    if ($1 != NULL)
        create_symbol_table_entry(@1.first_line, $1, entry_code);

    int inst = $2;
    append_inst(inst, NULL, NULL);

    sdsfree($1);
}
/* error handling for instructions */
            | maybe_label instructions_2_operands error '\n'
{
    int zerobased_token = insttoken_to_zero_index($2);
    sds s = sdscatprintf(sdsempty(),
                         "'%s' instruction needs two operands",
                         instructions_string[zerobased_token]);
    error(@2.first_line, s);
    sdsfree(s);
}
            | maybe_label instructions_1_operand error '\n'
{
    int zerobased_token = insttoken_to_zero_index($2);
    sds s = sdscatprintf(sdsempty(),
                         "'%s' instruction needs one operand",
                         instructions_string[zerobased_token]);
    error(@2.first_line, s);
    sdsfree(s);
}
            | maybe_label instructions_0_operand error '\n'
{
    int zerobased_token = insttoken_to_zero_index($2);
    sds s = sdscatprintf(sdsempty(),
                         "'%s' instruction needs no operands",
                         instructions_string[zerobased_token]);
    error(@2.first_line, s);
    sdsfree(s);
}
            ;

op_operand: OP_INT
{
    $$.kind = immediate;
    $$.value.i = $1;
}
          | OP_LABEL
{
    $$.kind = direct;
    $$.value.label[0] = $1;
}
          | '~' '(' OP_LABEL ',' OP_LABEL ')'
{
    $$.kind = dist;
    $$.value.label[0] = $3;
    $$.value.label[1] = $5;
}
          | REGISTER
{
    $$.kind = iregister;
    $$.value.i = $1;
}
;

directive_statement: maybe_label DIRECTIVE_DATA numbers '\n'
{
    if ($1 != NULL)
        create_symbol_table_entry(@1.first_line, $1, entry_data);

    int_array* il = $3;

    /* add the ints to the end of the data array */
    for (size_t i = 0; i < il->length; i++) {
        append_int_array(data, il->list[i]);
    }

    destroy_int_array(&$3);
}
                   | maybe_label DIRECTIVE_DATA error '\n'
{
    error(@2.first_line, "expected comma separated list of numbers after .data");
    sdsfree($1);
}
                   | maybe_label DIRECTIVE_STRING STRING '\n'
{
    if ($1 != NULL)
        create_symbol_table_entry(@1.first_line, $1, entry_data);

    sds s = $3;

    /* add the string to the end of the data array */
    for (size_t i = 0; i < sdslen(s); i++) {
        append_int_array(data, s[i]);
    }

    // null terminator
    append_int_array(data, 0);

    sdsfree(s);
}
                   | maybe_label DIRECTIVE_STRING error
{
    error(@2.first_line, "expected a quoted string after .string");
    sdsfree($1);
}
                   | maybe_label DIRECTIVE_ENTRY OP_LABEL '\n'
{
    symbol_table_entry* ste = find_symbol_table_entry($3);

    if (ste) {
        if (ste->type == entry_external) {
            sds s = sdscatprintf(sdsempty(),
                                 "label '%s' is defined as .extern, "
                                 "cannot also define it as .entry", $3);
            error(@3.first_line, s);
            sdsfree(s);
        }
        else
            ste->isentry = 1;
    }
    else {
        ste = make_symbol_table_entry();
        ste->label = sdsdup($3);
        ste->ic = ic();
        ste->type = entry_unknown; // we don't know yet
        ste->isentry = 1;
    }

    sdsfree($1);
    sdsfree($3);
}
                   | maybe_label DIRECTIVE_ENTRY error '\n'
{
    error(@2.first_line, "expected a label after .entry");
    sdsfree($1);
}
                   | maybe_label DIRECTIVE_EXTERN OP_LABEL '\n'
{
    create_symbol_table_entry(@3.first_line, $3, entry_external);
    sdsfree($1);
    sdsfree($3);
}
                   | maybe_label DIRECTIVE_EXTERN error '\n'
{
    error(@2.first_line, "expected a label after .extern");
    sdsfree($1);
}
;

numbers: INT
{
    $$ = make_int_array();
    append_int_array($$, $1);
}
       | numbers ',' INT
{
    append_int_array($1, $3);
}
       ;

%%

void destroy_inst_operand(inst_operand* op) {
    if (op->kind == direct) {
        sdsfree(op->value.label[0]);
    }
    else if (op->kind == dist) {
        sdsfree(op->value.label[0]);
        sdsfree(op->value.label[1]);
    }
}

void format_operand(int code_offset, int instruction_start_offset,
                    inst_operand op, int* v) {
    assert(op.kind != iregister);

    if (op.kind == immediate) {
        *v |= (op.value.i << 2);
    }
    else if (op.kind == direct) {
        append_replace_entry(code_offset, instruction_start_offset,
                             op.value.label[0], NULL);
        *v = ~0;
    }
    else if (op.kind == dist) {
        append_replace_entry(code_offset, instruction_start_offset,
                             op.value.label[0], op.value.label[1]);
        *v = ~0;
    }
}

void format_operands(int instruction_start_offset,
                     inst_operand* src,
                     inst_operand* dst) {
    assert(dst || !src);

    if (dst->kind == iregister || (src && src->kind == iregister)) {
        int* op = appendempty_int_array(code);

        if (src) {
            if (src->kind == iregister) {
                *op |= (src->value.i << 7);
            }
            else {
                format_operand(ic() - 1, instruction_start_offset, *src, op);
                op = appendempty_int_array(code);
            }
        }

        if (dst->kind == iregister) {
            *op |= (dst->value.i << 2);
        }
        else {
            if (src && src->kind == iregister) {
                op = appendempty_int_array(code);
            }

            format_operand(ic() - 1, instruction_start_offset, *dst, op);
        }

        return;
    }

    if (src) {
        int* srcop = appendempty_int_array(code);
        format_operand(ic() - 1, instruction_start_offset, *src, srcop);
    }

    int* dstop = appendempty_int_array(code);
    format_operand(ic() - 1, instruction_start_offset, *dst, dstop);
}

void append_inst(int type, inst_operand* op1, inst_operand* op2) {
    /* possible addressing values for each of the operands.
       first dimesion for src, second for dst. an entry for
       each possible value in enum operand_addressing */
    static int addressing[NUM_ADDRESSING][NUM_ADDRESSING] = {
        {0, 4, 8, 12}, // {0b000000, 0b000100, 0b001000, 0b001100},
        {16, 20, 24, 28}, // {0b010000, 0b010100, 0b011000, 0b011100},
        {32, 36, 40, 44}, // {0b100000, 0b100100, 0b101000, 0b101100},
        {48, 52, 56, 60}, // {0b110000, 0b110100, 0b111000, 0b111100}
    };

    int instruction_start_offset = ic();

    int opcount;

    /* the actual instruction */
    int* v = appendempty_int_array(code);

    /* opcode */
    *v |= (insttoken_to_zero_index(type) << 6);

    switch (type) {
        case INST_MOV:
        case INST_CMP:
        case INST_ADD:
        case INST_SUB:
        case INST_LEA:
            opcount = 2;
            break;
        case INST_NOT:
        case INST_CLR:
        case INST_INC:
        case INST_DEC:
        case INST_JMP:
        case INST_BNE:
        case INST_RED:
        case INST_PRN:
        case INST_JSR:
            opcount = 1;
            break;
        case INST_RTS:
        case INST_STOP:
            opcount = 0;
            break;
        default:
            assert(0 && "bad inst type");
    }

    if (opcount == 2) {
        assert(op1 && op2);
        inst_operand* src = op1;
        inst_operand* dst = op2;

        /* group */
        *v |= (2 << 10);

        /* addressing */
        *v |= addressing[src->kind][dst->kind];

        format_operands(instruction_start_offset, src, dst);
    }
    else if (opcount == 1) {
        assert(op2 == NULL);
        /* group */
        *v |= (1 << 10);

        /* addressing */
        *v |= addressing[0][op1->kind];

        format_operands(instruction_start_offset, NULL, op1);
    }
    else {
        assert(!op1 && !op2);
    }
}

/* create an entry in the symbol table (or update an unknown one) */
symbol_table_entry* create_symbol_table_entry(int line, sds label,
                                              symbol_table_entry_type type) {
    symbol_table_entry* entry = find_symbol_table_entry(label);

    /* if we found an entry that has a type, then we already saw it
     * and this is a redefinition */
    if (entry != NULL) {
        if (entry->type != entry_unknown) {
            sds s = sdscatprintf(sdsempty(),
                                 "redefinition of label '%s'",
                                 entry->label);
            error(line, s);
            sdsfree(s);

            return NULL;
        }
    }
    else {
        entry = make_symbol_table_entry();
        entry->label = sdsdup(label);
        entry->ic = ic();
    }

    entry->type = type;

    if (type == entry_code)
        entry->code_offset = ic();
    else if (type == entry_data)
        entry->code_offset = dc();

    return entry;
}

void compile_start() {
    assert(code == NULL);
    assert(data == NULL);
    assert(!saw_error);

    code = make_int_array();
    data = make_int_array();
}

void compile_cleanup() {
    destroy_int_array(&code);
    destroy_int_array(&data);
    destroy_symbol_table();

    saw_error = 0;
}

#ifndef LEXER_MAIN

static void print_logo() {
    printf("                  ___           ___           ___           ___     \n");
    printf("    ___          /  /\\         /  /\\         /  /\\         /  /\\    \n");
    printf("   /  /\\        /  /::\\       /  /:/        /  /:/        /  /:/    \n");
    printf("  /  /:/       /  /:/\\:\\     /  /:/ /\\     /  /:/        /  /:/     \n");
    printf(" /__/::\\      /  /:/~/::\\   /  /:/ /::\\   /  /:/  ___   /  /:/  ___ \n");
    printf(" \\__\\/\\:\\__  /__/:/ /:/\\:\\ /__/:/ /:/\\:\\ /__/:/  /  /\\ /__/:/  /  /\\\n");
    printf("    \\  \\:\\/\\ \\  \\:\\/:/__\\/ \\  \\:\\/:/~/:/ \\  \\:\\ /  /:/ \\  \\:\\ /  /:/\n");
    printf("     \\__\\::/  \\  \\::/       \\  \\::/ /:/   \\  \\:\\  /:/   \\  \\:\\  /:/ \n");
    printf("     /__/:/    \\  \\:\\        \\__\\/ /:/     \\  \\:\\/:/     \\  \\:\\/:/  \n");
    printf("     \\__\\/      \\  \\:\\         /__/:/       \\  \\::/       \\  \\::/   \n");
    printf("                 \\__\\/         \\__\\/         \\__\\/         \\__\\/    \n");
    printf("\n                   Idan's AS Compiler Collection\n");
}

int main(int argc, const char* argv[]) {
    if (argc == 1) {
        print_logo();
        printf("\n\n");
        printf("usage: %s <file>...\n", argv[0]);
        return 1;
    }

    int rc = 0;

    for (int i = 1; i < argc; i++) {
        sds infilename, outfilename, entfilename, extfilename;
        FILE *asfile = NULL;

        if (strcmp(argv[i], "-") == 0) {
            asfile = stdin;

            infilename = sdsnew("<stdin>");
            outfilename = sdsnew("stdin.o");
            entfilename = sdsnew("stdin.ent");
            extfilename = sdsnew("stdin.ext");
        }
        else {
            infilename = sdsnew(argv[i]);
            infilename = sdscat(infilename, ".as");
            asfile = fopen(infilename, "r");

            outfilename = sdscatprintf(sdsempty(), "%s.o", argv[i]);
            entfilename = sdscatprintf(sdsempty(), "%s.ent", argv[i]);
            extfilename = sdscatprintf(sdsempty(), "%s.ext", argv[i]);

            if (!asfile) {
                rc = 1;

                printf("error opening '%s', skipping\n", infilename);

                sdsfree(infilename);
                sdsfree(outfilename);
                sdsfree(entfilename);
                sdsfree(extfilename);

                continue;
            }
        }

        yyin = asfile;
        compile_start();

        do {
            yyparse();
        } while (!feof(yyin));

        // try to resolve labels regardless if we saw errors, so we'll
        // warn for missing labels
        resolve_replacements();

        /* don't produce output files if we saw an error */
        if (!saw_error) {
            /* write the code and data to the object file  */
            FILE* outfile = fopen(outfilename, "wb");
            if (!outfile) {
                rc = 1;

                sds s = sdscatprintf(sdsempty(),
                                     "error opening %s",
                                     outfilename);
                perror(s);
                sdsfree(s);
            }
            else {
                fprintf(outfile, "%x %x\n", ic(), dc());

                for (unsigned int i = 0; i < code->length; i++) {
                    assert(code->list[i] < 0xfff);
                    int inst = code->list[i] & 0xfff;
                    fprintf(outfile, "%02x %03x\n", 100 + i, inst);
                }

                for (unsigned int i = 0; i < data->length; i++) {
                    assert(data->list[i] < 0xfff);
                    int d = data->list[i] & 0xfff;
                    fprintf(outfile, "%02x %03x\n", 100 + ic() + i, d);
                }

                fclose(outfile);

                /* .ext and .ent files */
                rc |= write_external_entry_files(extfilename, entfilename);
            }
        }
        else {
            printf("no code generated for %s, found %d error(s)\n",
                   infilename, saw_error);

            rc = 1;
        }

        compile_cleanup();

        sdsfree(infilename);
        sdsfree(outfilename);
        sdsfree(entfilename);
        sdsfree(extfilename);
    }

    return rc;
}

#endif

void yyerror(const char* s) {
    s = s;
}

void error(int line, const char *s) {
	printf("syntax error, line %d: %s\n", line, s);
    saw_error++;
}
